/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _APE_
#define _APE_

#include "ape.hpp"

#include "../pugixml/pugixml.hpp"

#include <set>
#include <vector>
#include <string>
#include <sstream>
#include <cstring>
#include <wordexp.h>
#include <fstream>
#include <string>
#include <cerrno>
#include <iterator>
#include <iostream>

ape::tree::tree() {
}

ape::tree::tree(const char* source, const char* root, bool file, bool cin) {
	parse(source, root, file, cin);
}

pugi::xml_node ape::tree::parse(const char* source, const char* root, bool file, bool cin) {
	xml = file ? get_file_contents(source, cin) : source;
	result = doc.load(xml.c_str());
	if (!result) {
		std::string error;
		error += result.description();
		if (result.offset) {
			std::string what = xml.c_str()+result.offset;
			what = what.substr(0, what.find("\n")-1);
			std::string position = ((std::ostringstream&)(std::ostringstream() << result.offset)).str();
			error += (std::string)" at character "+position+" [..."+what+"...]";
		}
		throw error;
	}
	if (root and !doc.child(root))
		throw (std::string)"No root node named '"+root+"' found";
	return doc.first_child();
}

std::string ape::tree::get_file_contents(std::string file, bool cin) {
	if (cin and (file == "" or file == "-")) {
		std::cin >> std::noskipws;
		return std::string(std::istream_iterator<char>(std::cin), std::istream_iterator<char>());
	}
	std::ifstream in(file.c_str(), std::ios::in | std::ios::binary);
	if (in) {
		std::string contents;
		in.seekg(0, std::ios::end);
		contents.resize(in.tellg());
		in.seekg(0, std::ios::beg);
		in.read(&contents[0], contents.size());
		in.close();
		return contents;
	}
	throw (std::string)"Cannot get file contents of '"+file+"': "+strerror(errno);
}

std::vector<std::string> ape::explore(pugi::xml_node node) {
	std::vector<std::string> steps;
	do {
		std::string step = node.name();
		size_t position = 0;
		for (pugi::xml_node cursor = node; cursor; cursor = cursor.previous_sibling(step.c_str()))
			++position;
		if (position > 1)
			step += "["+((static_cast<std::ostringstream&>(std::ostringstream().flush() << position))).str()+"]";
		steps.insert(steps.begin(), step);
	}
	while ((node = node.parent()));
	return steps;
}

std::string ape::implode(std::vector<std::string> subject) {
	std::string result = subject[0];
	for (std::vector<std::string>::iterator it = subject.begin()+1; it != subject.end(); ++it)
		result += "/"+*it;
	return result;
}

ssize_t ape::level(const pugi::xml_node* subject, const pugi::xml_node* reference, bool* related) {
	if (!*subject or !*reference)
		throw "Cannot analyse null nodes.";
	std::vector<std::string> subject_steps = explore(*subject);
	if (!reference)
		return subject_steps.size()-1;
	std::vector<std::string> reference_steps = explore(*reference);
	size_t size = subject_steps.size();
	if (size > reference_steps.size())
		size = reference_steps.size();
	if (related) {
		*related = true;
		for (size_t i = 1; i < size and *related; ++i)
			if (subject_steps[i] != reference_steps[i])
				*related = false;
	}
	return subject_steps.size()-reference_steps.size();
}

bool ape::has_ancestor(pugi::xml_node subject, pugi::xml_node reference, ssize_t offset) {
	bool related;
	ssize_t generations = level(&subject, &reference, &related);
	return related and generations > -offset;
}

std::vector<std::string> ape::perform_word_expansion(std::string word, int flags) {
	wordexp_t p;
	const char* error;
	switch (wordexp(word.c_str(), &p, flags)) {
		case WRDE_BADCHAR: error = "Illegal occurrence of newline or one of |, &, ;, <, >, (, ), {, }."; break;
		case WRDE_BADVAL:  error = "An undefined shell variable was referenced, and the WRDE_UNDEF flag told us to consider this an error."; break;
		case WRDE_CMDSUB:  error = "Command substitution occurred, and the WRDE_NOCMD flag told us to consider this an error."; break;
		case WRDE_NOSPACE: error = "Out of memory."; break;
		case WRDE_SYNTAX:  error = "Shell syntax error, such as unbalanced parentheses or unmatched quotes."; break;
		case 0:            error = 0; break;
		default:           error = "An unknwon error occured."; break;
	}
	if (error)
		throw error;
	std::vector<std::string> words;
	for (size_t i = 0; i < p.we_wordc; ++i)
		words.push_back(p.we_wordv[i]);
  	wordfree(&p);
  	return words;
}

void ape::resolve_fpath(pugi::xml_node node, const char* root, const char* tag) {
	typedef std::set<std::string> trace_t;
	static trace_t trace;
	if (!strcmp(node.name(), tag)) {
		std::vector<std::string> words = perform_word_expansion(node.text().get());
		for (std::vector<std::string>::iterator word = words.begin(); word != words.end(); ++word) {
			std::pair<trace_t::iterator, bool> ret = trace.insert(*word);
			tree dst;
			pugi::xml_node file;
			if (!ret.second) {
				trace.erase(ret.first);
				throw (std::string)"Prevented endless recursion in file '"+ret.first->c_str()+"'";
			}
			else {
				try {
					file = dst.parse(ret.first->c_str(), root);
				}
				catch (std::string& error) {
					trace.erase(ret.first);
					throw error;
				}
			}
			for (pugi::xml_node n = file.first_child(); n and n.type() == 2; n = n.next_sibling())
				resolve_fpath(node.parent().insert_copy_before(n, node), root, tag);
			trace.erase(ret.first);
		}
		node.parent().remove_child(node);
		return;
	}
	for (pugi::xml_node child = node.first_child(); child and child.type() == 2; child = child.next_sibling())
		resolve_fpath(child, root, tag);
}

void ape::resolve_xpath(pugi::xml_node node, const char* tag) {
	typedef std::vector<pugi::xml_node> trace_t;
	static trace_t trace;
	if (!strcmp(node.name(), tag)) {
		pugi::xpath_node_set node_set;
		try {
			node_set = node.select_nodes(node.text().get());
		}
		catch (pugi::xpath_exception& exception) {
			throw (std::string)exception.what()+" in xpath '"+node.text().get()+"'";
		}
		if (!node_set.size())
			throw (std::string)"No matches for xpath '"+node.text().get()+"'";
		trace.push_back(node);
		for (pugi::xpath_node_set::const_iterator n = node_set.begin(); n != node_set.end(); ++n) {
			for (trace_t::iterator t = trace.begin(); t != trace.end(); ++t)
				if (has_ancestor(*t, n->node(), 1)) {
					trace.pop_back();
					throw (std::string)"Prevented endless recursion at xpath '"+node.text().get()+"'";
				}
			if (!strcmp(n->node().name(), tag))
				throw (std::string)"Missing container node for resolving xpath '"+n->node().text().get()+"' caused by xpath '"+node.text().get()+"' directly pointing to it";
			resolve_xpath(n->node(), tag);
			node.parent().insert_copy_before(n->node(), node);
		}
		trace.pop_back();
		node.parent().remove_child(node);
		return;
	}
	for (pugi::xml_node child = node.first_child(); child and child.type() == 2; child = child.next_sibling())
		resolve_xpath(child, tag);
}

void ape::merge_twins(pugi::xml_node node) {
	for (pugi::xml_node child = node.last_child(); child and child.type() == 2; child = child.previous_sibling()) {
		for (pugi::xml_node twin = child.previous_sibling(child.name()); twin and twin.type() == 2;) {
			for (pugi::xml_node twins_child = twin.last_child(); twins_child and twins_child.type() == 2; twins_child = twins_child.previous_sibling())
				if (child.first_child())
					child.insert_copy_before(twins_child, child.first_child());
				else
					child.append_copy(twins_child);
			pugi::xml_node cloned_twin = twin;
			twin = twin.previous_sibling(child.name());
			node.remove_child(cloned_twin);
		}
		merge_twins(child);
	}
}

void ape::inherit_attributes(pugi::xml_node node) {
	for (pugi::xml_node child = node.first_child(); child and child.type() == 2; child = child.next_sibling()) {
		for (pugi::xml_attribute attr = node.first_attribute(); attr; attr = attr.next_attribute()) {
			if (!child.attribute(attr.name()))
				child.append_copy(attr);
		}
		if (child.first_child() and child.first_child().type() == 2)
			inherit_attributes(child);
	}
}

#endif
