/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PIBOX_MIXED_
#define _PIBOX_MIXED_

#include "mixed.hpp"

pibox::mixed::mixed() {
	size = 0;
	ram_begin = 0;
}

pibox::mixed::mixed(size_t item_size, size_t data_size) {
	size = 0;
	reserve(item_size, data_size);
}

pibox::mixed::~mixed() {
	delete[] ram_begin;
}

void pibox::mixed::reserve(size_t item_size, size_t data_size) {
	item_box.reserve(item_size);
	char* new_ram_begin = 0;
	try {
		new_ram_begin = new char[data_size];
	}
	catch (std::bad_alloc& error) {
		throw (std::string)"No RAM left to reserve "+to_string(data_size)+" B";
	}
	ram_end = new_ram_begin+data_size;
	if (size) {
		size_t left = data_size;
		for (size_t i = 0; i < item_box.used(); ++i) {
			item_t& item = item_box.front(i);
			if (item.size > left)
				item_box.pop_back();
			else {
				for (size_t e = data_size-left; e < item.size; ++e)
					new_ram_begin[e] = item.data[e];
				item.data = ram_end-left;
				left -= item.size;
			}
		}
		data_begin = ram_begin = new_ram_begin;
		size = data_size-left;
		data_end = ram_begin+size;
	}
	else {
		data_begin = ram_end;
		data_end = ram_begin = new_ram_begin;
	}
}

pibox::mixed::item_t const& pibox::mixed::make_front(size_t size) {
	item_t new_front = {size, data_begin-size};
	if (data_begin < data_end) {
		if (size > (size_t)(data_begin-ram_begin)) {
			new_front.data = ram_end-size;
			if ((size_t)(ram_end-data_end) < size)
				throw (std::string)"No space left to push data front, jumped";
		}
	}
	else {
		if (size > (size_t)(data_begin-data_end))
			throw (std::string)"No space left to push data front";
	}
	item_box.push_front(new_front);
	data_begin = new_front.data;
	this->size += size;
	return item_box.front();
}

void pibox::mixed::push_front(const void* data, size_t size) {
	copy(make_front(size).data, data, size);
}

pibox::mixed::item_t pibox::mixed::shrink_front(size_t size) {
	item_t& front = item_box.front();
#if PIBOX_ENABLE_SAFE_MODE
	if (!(size < front.size))
		throw (std::string)"Invalid size to shrink front to";
#endif
	item_t old_front = {size, front.data};
	front.size -= size;
	front.data = data_begin += size;
	this->size -= size;
	return old_front;
}

pibox::mixed::item_t pibox::mixed::stretch_front(size_t size) {
	item_t new_front = {size, data_begin-size};
	if (data_begin < data_end) {
		if (size > (size_t)(data_begin-ram_begin))
			throw (std::string)"No space left to stretch data front, would jump";
	}
	else {
		if (size > (size_t)(data_begin-data_end))
			throw (std::string)"No space left to stretch data front";
	}
	item_t& front = item_box.front();
	front.size += size;
	front.data = data_begin = new_front.data;
	this->size += size;
	return new_front;
}

void pibox::mixed::prepend_front(const void* data, size_t size) {
	copy(stretch_front(size).data, data, size);
}

pibox::mixed::item_t& pibox::mixed::fuse_front() {
	item_t* front = &item_box.front(), * new_front = front;
	item_box.increment(new_front);
	if (front->data > new_front->data)
		throw (std::string)"No front to fuse, would jump";
	new_front->data = front->data;
	new_front->size += front->size;
	item_box.pop_front();
	return this->front();
}

pibox::mixed::item_t const& pibox::mixed::make_back(size_t size) {
	item_t new_back = {size, data_end};
	if (data_begin < data_end) {
		if (size > (size_t)(ram_end-data_end)) {
			new_back.data = ram_begin;
			if ((size_t)(data_begin-ram_begin) < size)
				throw (std::string)"No space left to push data back, jumped";
		}
	}
	else {
		if (size > (size_t)(data_begin-data_end))
			throw (std::string)"No space left to push data back";
	}
	item_box.push_back(new_back);
	data_end = new_back.data+size;
	this->size += size;
	return item_box.back();
}

void pibox::mixed::push_back(const void* data, size_t size) {
	copy(make_back(size).data, data, size);
}

pibox::mixed::item_t pibox::mixed::shrink_back(size_t size) {
	item_t& back = item_box.back();
#if PIBOX_ENABLE_SAFE_MODE
	if (!(size < back.size))
		throw (std::string)"Invalid size to shrink back to";
#endif
	back.size -= size;
	data_end -= size;
	item_t old_back = {size, data_end};
	this->size -= size;
	return old_back;
}

pibox::mixed::item_t pibox::mixed::stretch_back(size_t size) {
	item_t new_back = {size, data_end};
	if (data_begin < data_end) {
		if (size > (size_t)(ram_end-data_end))
			throw (std::string)"No space left to stretch data back, would jump";
	}
	else {
		if (size > (size_t)(data_begin-data_end))
			throw (std::string)"No space left to stretch data back";
	}
	item_box.back().size += size;
	data_end += size;
	this->size += size;
	return new_back;
}

void pibox::mixed::append_back(const void* data, size_t size) {
	copy(stretch_back(size).data, data, size);
}

pibox::mixed::item_t& pibox::mixed::fuse_back() {
	item_t* back = &item_box.back(), * new_back = back;
	item_box.decrement(new_back);
	if (back->data < new_back->data)
		throw (std::string)"No back to fuse, would jump";
	new_back->data = back->data;
	new_back->size += back->size;
	item_box.pop_back();
	return this->back();
}

void pibox::mixed::pop_front() {
	size -= item_box.front().size;
	item_box.pop_front();
	if (size)
		data_begin = item_box.front().data;
	else {
		data_begin = ram_end;
		data_end = ram_begin;
	}
}

void pibox::mixed::pop_back() {
	size -= item_box.back().size;
	item_box.pop_back();
	if (size) {
		item_t& back = item_box.back();
		data_end = back.data+back.size;
	}
	else {
		data_begin = ram_end;
		data_end = ram_begin;
	}
}

void pibox::mixed::data_left(size_t& front, size_t& back) {
	if (data_begin < data_end) {
		front = data_begin-ram_begin;
		back = ram_end-data_end;
	}
	else
		front = back = data_begin-data_end;
}

void pibox::mixed::clear() {
	size = 0;
	data_begin = ram_end;
	data_end = ram_begin;
	item_box.clear();
}

void pibox::mixed::free() {
	delete[] ram_begin;
	ram_begin = 0;
	ram_end = 0;
	clear();
	item_box.free();
}

#endif
