/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PICONF_DAQ_TAG_
#define _PICONF_DAQ_TAG_

#include "daq_tag.hpp"
#include <sstream>
#include <arpa/inet.h>

piconf::daq_tag::file_t::file_t() {
	state = 0;
	data = 0;
}

piconf::daq_tag::ring_t::ring_t() {
	ring = 0;
	item_done = data_done = 0;
}

piconf::daq_tag::daq_tag(pugi::xml_node node)
: data_tag(node) {
}

void piconf::daq_tag::setup() {
	daq_state = 0;
	item_done = 0;
	data_done = 0;
	if (node.child("file")) {
		daq_state |= DAQ;
		item_size = node.attribute("item_size").as_uint();
		data_size = node.attribute("data_size").as_uint();
		if (!data_size)
			data_size = DEF_DATA_SIZE;
		if (data_size > MAX_DATA_SIZE)
			throw (std::string)"Size of data buffer "+calculate_size(data_size)+" exceeds maximum of "+calculate_size(MAX_DATA_SIZE);
		if (!item_size)
			item_size = DEF_ITEM_SIZE;
		if (item_size > MAX_ITEM_SIZE)
			throw (std::string)"Size of item buffer "+to_string(item_size)+" exceeds maximum of "+calculate_size(MAX_ITEM_SIZE);
		size_t id_raw = 0;
		std::string id_str = to_string(id_raw);
		for (pugi::xml_node child = node.first_child(); child and child.type() == 2; child = child.next_sibling()) {
			while (VIO::File::exists(file(id_str, child).c_str()))
				if (0 == ++id_raw)
					throw (std::string)"Entropy of DAQ IDs exhausted";
				else
					id_str = to_string(id_raw);
		}
		for (pugi::xml_node child = node.first_child(); child and child.type() == 2; child = child.next_sibling()) {
			file_box.push_back(file_t());
			file_box.back().data = 0;
			file_box.back().file.assign(file(id_str, child).c_str());
			try {
				file_box.back().file.mkdir();
			}
			catch (VIO::IO::Error& error) {
			}
			try {
				file_box.back().file.open(O_RDWR | O_CREAT);
			}
			catch (VIO::IO::Error& error) {
				piconf::error.set(error_t::FILE_IO_ERROR | error_t::PICONFM_ERROR);
				throw (std::string)error.what();
			}
		}	
	}
	else {
		//only status readback OR daq command
	}
	pugi::xml_attribute readback = node.attribute("readback");
	if (readback) {
		daq_state |= READBACK;
		rb.path = readback.value();
		if (rb.path.size()) {
			daq_state |= FILE_RB;
			rb.name = node.attribute("name").value();
		}
	}
	data_tag::setup();
}

bool piconf::daq_tag::instruct(io_tag* io, bool fail, bool exit) {
	data_tag::instruct(io, fail, exit);
	if (io_box[io].state & io_t::EXIT)
		return false;
	else
	if (io_box[io].state & io_t::PARSED_RESPONSE) {
		if (io_box[io].state & io_t::VALID_RESPONSE) {
			if (daq_state & DAQ) {
				uint32_t size = ntohl(ring_map[io].head[0]);
				ring_map[io].daq_state = ntohl(ring_map[io].head[1]);
				ring_map[io].s.item_used = ntohl(ring_map[io].head[2]);
				ring_map[io].s.data_used = ntohl(ring_map[io].head[3]);
				if (size) {
					daq_state &= ~IDLE;
					//shrink_front... before making next one
					ring_map[io].ring->make_front(size);
					ring_map[io].i_item = ring_map[io].ring->rbegin();
					io->fetch(ring_map[io].i_item->data, size);
					//separate data and head fetches concerning instruct call?
				}
				else {
					daq_state |= IDLE;
				}
				io->fetch(&ring_map[io].head, sizeof ring_map[io].head);
				return true;
			}
		}
		return false;
		//re-send via?!:
		//	state = 0;
		//	data_tag::instruct(io, fail);
	}
	return true;
}

void piconf::daq_tag::parse_response(io_tag* io, pugi::xml_node node) {
	data_tag::parse_response(io, node);
	std::string error;
	if (node.attribute("error")) {
		error = node.text().get();
		daq_state |= FAIL;
	}
	else {
		io_box[io].state |= io_t::VALID_RESPONSE;
		io_box[io].size = htonl(0);
		if (daq_state & READBACK) {
			if (daq_state & FILE_RB) {
				file_map.insert(std::make_pair(io, file_t()));
				std::ostringstream os;
				node.print(os);
				rb.data = os.str();
				file_map[io].state = file_t::DONE;
				file_map[io].data = &rb.data;
				file_map[io].file.assign((rb.path+"/"+io->id()+"/"+rb.name).c_str());
			}
			else {
				for (pugi::xml_node child = node.first_child(); child and child.type() == 2; child = child.next_sibling())
					io_box[io].result.push_back((std::string)"   "+child.path().c_str()+": "+child.text().get());
			}
		}
		if (daq_state & DAQ) {
			ring_map.insert(std::make_pair(io, ring_t()));
			ring_map[io].ring = new pibox::mixed(item_size, data_size);
			ring_map[io].m.item_size = item_size;
			ring_map[io].m.data_size = data_size;
			ring_map[io].s.item_size = node.attribute("item_size").as_uint();
			ring_map[io].s.data_size = node.attribute("data_size").as_uint();
			ring_map[io].head[0] = htonl(0);
			ring_map[io].head[1] = htonl(0);
			ring_map[io].head[2] = htonl(node.attribute("item_used").as_uint());
			ring_map[io].head[3] = htonl(node.attribute("data_used").as_uint());
		}
	}
	if (error.size())
		io_box[io].result.push_back((std::string)"   "+node.path().c_str()+": "+error);
}

bool piconf::daq_tag::process() {
	bool undone = false;
	if (!data_tag::process()) {
		if (!(daq_state & FAIL)) {
			for (file_map_t::iterator it = file_map.begin(); it != file_map.end(); ++it) {
				file_t* file = &it->second;
				if (!(file->state & file_t::DONE)) {
					if (file->state & file_t::WRITING) {
						if (!file->file.hasToWrite()) {
							file->state |= file_t::DONE;
						}
					}
					else {
						try {
							file->file.remove();
						}
						catch (VIO::IO::Error& error) {
						}
						try {
							file->file.mkdir();
						}
						catch (VIO::IO::Error& error) {
						}
						try {
							file->file.open(O_RDWR | O_CREAT);
						}
						catch (VIO::IO::Error& error) {
							io_box[it->first].result.push_back("   /daq: "+(std::string)error.what());
							file->state |= file_t::DONE;
							piconf::error.set(error_t::FILE_IO_ERROR | error_t::PICONFM_ERROR);
						}
						file->file.push(file->data->c_str(), file->data->size());
						file->state |= file_t::WRITING;
					}
					if (!(file->state & file_t::DONE))
						undone = true;
				}
			}
			bool none = true;
			for (io_box_t::iterator it = io_box.begin(); it != io_box.end(); ++it) {
				if (it->second.state & io_t::EXIT)
					continue;
				else
					none = false;
				if (it->second.state & io_t::VALID_RESPONSE) {
					if (daq_state & DAQ) {
						undone = true;
						ring_t& ring = ring_map[it->first];
						ring.m.item_used = ring.ring->item_used();
						ring.m.data_used = ring.ring->data_used();
						it->second.result.clear();
						it->second.result.push_back("   Acquired "+to_string(ring.item_done)+" items ("+calculate_size(ring.data_done)+"):");
						it->second.result.push_back("    "+calculate_percentage(ring.m.item_used, ring.m.item_size)+" ("+to_string(ring.m.item_used)+" items), "+calculate_percentage(ring.m.data_used, ring.m.data_size)+" ("+calculate_size(ring.m.data_used)+") locally queued.");
						it->second.result.push_back("    "+calculate_percentage(ring.s.item_used, ring.s.item_size)+" ("+to_string(ring.s.item_used)+" items), "+calculate_percentage(ring.s.data_used, ring.s.data_size)+" ("+calculate_size(ring.s.data_used)+") remotely queued.");
					}
					else {
					}
				}
				else
					undone = true;
				//if (!it->second.result.size())
				//	it->second.result.push_back("   bdbdbd");
			}
			if (none)
				undone = false;
		}
	}
	else
		undone = true;
	if (!undone and io_box.size())
		daq_state |= DONE;
	else
		daq_state &= ~DONE;
	return undone;
}

void piconf::daq_tag::status(lines_t& lines) {
	std::string status;
	if (daq_state & FAIL)
		status = "failed";
	else
	if (daq_state & DONE)
		status = "done";
	else
	if (daq_state & IDLE)
		status = "idle";
	else
		status = "fine";
	if (daq_state & DAQ) {
		lines.push_back(" Processing DAQ:");
		lines.push_back((std::string)"  Totally acquired "+to_string(item_done)+" items ("+calculate_size(data_done)+"), "+status+".");
	}
	else {
		lines.push_back(" Investigating DAQ: "+status+".");
	}
}

piconf::daq_tag::~daq_tag() {
	for (ring_map_t::iterator ring = ring_map.begin(); ring != ring_map.end(); ++ring)
		delete ring->second.ring;
}
#endif
