/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PICONF_DATA_TAG_
#define _PICONF_DATA_TAG_

#include "../piconfm.hpp"
#include "../../../ape/ape.hpp"
#include <vector>
#include <string>
#include <arpa/inet.h>

piconf::data_tag::io_t::io_t() {
	state = 0;
	size = htonl(0);
}

piconf::data_tag::data_tag(pugi::xml_node node) 
: tag(node) {
}

void piconf::data_tag::subscribe(std::deque<io_tag*>& io_box, bool inherit) {
	if (inherit)
		for (std::deque<io_tag*>::iterator io = io_box.begin(); io != io_box.end(); ++io)
			(*io)->subscribe(this);
	else
	if (io_box.size())
		io_box.back()->subscribe(this);
}

bool piconf::data_tag::instruct(io_tag* io, bool fail, bool exit) {
	if (io_box.find(io) == io_box.end()) {
		io_box.insert(std::make_pair(io, io_t()));
	}
	if (exit) {
		io_box[io].state |= io_t::EXIT;
		return false;
	}
	if (fail)
		io_box[io].state = 0;
	if (io_box[io].state & io_t::PARSED_RESPONSE) {
		return false;
	}
	else
	if (io_box[io].state & io_t::INSTRUCTED_RESPONSE_XML) {
		io_box[io].state |= io_t::RECEIVED_RESPONSE;
		/* TODO
		 *
		 * This instruct method should only care about PEER_IO states
		 * and not data processing like PARSED_RESPONSE. Instead we
		 * should return false here and move the response parsing
		 * from process() into parse_result() and this will be called
		 * here. This would save time, since for now we sometimes loose
		 * some ms when we go into VIO::IO::process() since in some
		 * cases we didn't instructed anything... we just won't hang
		 * there forever because we have a periodic timer awaiking
		 * us every 10ms.
		 */
	}
	else
	if (io_box[io].state & io_t::INSTRUCTED_RESPONSE_SIZE) {
		if (!io_box[io].data_box.size()) {
			size_t size = ntohl(io_box[io].size);
			if (size > MAX_XML_SIZE) {
				piconf::error.set(error_t::MAXIMUM_ERROR | error_t::PICONFM_ERROR);
				throw (std::string)"XML size of "+calculate_size(size)+" exceeds maximum of "+calculate_size(MAX_XML_SIZE);
			}
			io_box[io].data_box.push_back(std::make_pair(new char[size], size));
		}
		io->fetch(io_box[io].data_box[0].first, io_box[io].data_box[0].second);
		io_box[io].state |= io_t::INSTRUCTED_RESPONSE_XML;
	}
	else {
		//io->push(&io_box[io].size, sizeof io_box[io].size);
		//io->push(xml.c_str(), xml.size()+1);
		io->push(query, (sizeof size)+xml.size()+1);
		io->fetch(&io_box[io].size, sizeof io_box[io].size);
		io_box[io].state |= io_t::INSTRUCTED_RESPONSE_SIZE;
	}
	return true;
}

void piconf::data_tag::result(io_tag* io, lines_t& lines) {
	if (io_box.find(io) != io_box.end())
		lines.insert(lines.end(), io_box[io].result.begin(), io_box[io].result.end());
}

void piconf::data_tag::setup() {
	save();
	tag::setup();
}

bool piconf::data_tag::process() {
	bool undone = false;
	for (io_box_t::iterator it = io_box.begin(); it != io_box.end(); ++it) {
		if (it->second.state & io_t::EXIT)
			continue;
		if (!(it->second.state & io_t::PARSED_RESPONSE)) {
			if (it->second.state & io_t::RECEIVED_RESPONSE) {
				std::string error;
				if (it->second.data_box[0].first[it->second.data_box[0].second-1] != '\0') {
					error = "Received non null terminated response";
					piconf::error.set(error_t::PARSING_ERROR | error_t::PICONFS_ERROR);
				}
				ape::tree tree;
				pugi::xml_node root;
				try {
					root = tree.parse(it->second.data_box[0].first, 0, false);
				}
				catch (std::string& details) {
					error = "Received invalid XML response: "+details;
					piconf::error.set(error_t::PARSING_ERROR | error_t::PICONFS_ERROR);
				}
				if (error.size())
					it->second.result.push_back("   /: "+error);
				else
					parse_response(it->first, root);
				it->second.state |= io_t::PARSED_RESPONSE;
			}
			else
				undone = true;
		}
	}
	if (tag::process())
		undone = true;
	return undone;
}

void piconf::data_tag::parse_response(io_tag* io, pugi::xml_node node) {
	int error = node.attribute("error").as_int();
	piconf::error.set(error);
}

bool piconf::data_tag::complete() {
	return true;
}

std::string piconf::data_tag::root() {
	return node.name();
}

piconf::data_tag::~data_tag() {
	for (io_box_t::iterator it = io_box.begin(); it != io_box.end(); ++it)
		for (size_t e = 0; e < it->second.data_box.size(); ++e)
			delete[] it->second.data_box[e].first;
}

#endif
